#!C:\bld\reprounzip-qt_1531157594865\_h_env\pythonw.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'reprounzip-qt==1.0.13','gui_scripts','reprounzip-qt'
__requires__ = 'reprounzip-qt==1.0.13'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('reprounzip-qt==1.0.13', 'gui_scripts', 'reprounzip-qt')()
    )
